#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import os

import psutil

from flask import current_app
import json

from ssk.globals.api_gate import ApiGate
from ssk.utils import get_safe_string


class ApiHandler(object):
    API_ERROR_TEMPLATE = "%s key %s version %s code %s message %s"

    @staticmethod
    def status(a_version, a_key):
        my_response = ApiGate.TOO_BUSY
        my_mesg = ""
        my_ret_val = 400

        if ApiGate.is_active_now(a_key):
            my_mesg = ApiGate.DONE
            my_ret_val = 200
        elif ApiGate.num_active_keys() < current_app.config["MAX_ACTIVE_KEYS"]:
            # this call works even if api is closed
            if ApiGate.is_valid(current_app, a_key, ignore_open=True):
                my_mesg = ApiGate.DONE
                my_ret_val = 200
            else:
                my_mesg = ApiGate.INVALID_KEY

        if my_ret_val == 200:
            my_process = psutil.Process(os.getpid())
            my_mem = round(my_process.memory_info().rss / (1024 ** 2), 2)
            my_data = {"mem": my_mem}

            my_response = json.dumps(my_data)
        else:
            current_app.logger.error(ApiHandler.API_ERROR_TEMPLATE, "status",
                                     get_safe_string(a_key),
                                     get_safe_string(a_version),
                                     my_ret_val,
                                     get_safe_string(my_mesg))

        return my_response, my_ret_val

    @staticmethod
    def echo(a_request, a_version, a_key):
        my_response = ApiGate.TOO_BUSY
        my_mesg = ""
        my_ret_val = 400

        if ApiGate.is_active_now(a_key):
            my_mesg = ApiGate.DONE
            my_ret_val = 200
        elif ApiGate.num_active_keys() < current_app.config["MAX_ACTIVE_KEYS"]:
            if ApiGate.is_valid(current_app, a_key):
                my_mesg = ApiGate.DONE
                my_ret_val = 200
            else:
                my_mesg = ApiGate.INVALID_KEY

        if my_ret_val == 200:
            my_content = a_request.json

            my_response = "I can't hear you"
            if "echo" in my_content.keys():
                my_response = get_safe_string(my_content["echo"])
        else:
            current_app.logger.error(ApiHandler.API_ERROR_TEMPLATE, "echo",
                                     get_safe_string(a_key),
                                     get_safe_string(a_version),
                                     my_ret_val,
                                     get_safe_string(my_mesg))

        return my_response, my_ret_val



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from .admin_cmd import AdminCmd
from .abstract_cmd import AbstractCmd
from flask_login import current_user

from ...db import get_db
from ...models.audit import Audit


class RootCmd(AbstractCmd):
    def __init__(self, a_name="."):
        super().__init__(a_name)
        self.reg_cmd(["a", "admin"], AdminCmd())

    def action(self, a_params):
        my_ret_val = False
        my_ret_mesg = ""

        my_cmd_name = a_params.pop(0)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_ret_val, my_ret_mesg = my_cmd.exec(a_params)

            my_audit = Audit()
            my_audit.category = "CMD"
            my_audit.by_user = current_user.email
            my_audit.description = "{} {}: {}".format(my_cmd_name, " ".join(a_params), my_ret_val)

            if my_ret_val:
                my_audit.status = "OK"
            else:
                my_audit.status = "NOK"

            get_db().session.add(my_audit)
            get_db().session.commit()

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        my_names = []
        my_mesg = '[[ print "'
        for my_tmp_cmd in self.get_all_cmds().values():
            my_name = my_tmp_cmd.get_name()
            if my_name not in my_names:
                my_mesg += "\n{}".format(my_name)
                my_names.append(my_name)

        my_mesg += '" ]]'

        return my_mesg



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import os
import sys
from datetime import datetime

import math
from flask import current_app
from math import floor
from ssk.globals.setting_parser import SettingParser
from markupsafe import escape


def get_safe_string(a_string):
    my_ret_val = ""

    if " " in a_string or a_string.isalnum():
        my_ret_val = escape(a_string)

    return str(my_ret_val)


def set_const(an_env, a_def=None, a_type=SettingParser.STR_TYPE):
    my_ret_val = os.getenv(an_env, None)
    if my_ret_val is None:
        if a_def is None:
            print("{} not found".format(an_env))
            sys.exit(1)
        else:
            #print("{} env var not found, setting from config file {}".format(an_env, a_def))
            my_ret_val = a_def
    else:
        print("{} env var overwrites config file value".format(an_env))

    my_ret_val = SettingParser.parse(a_type, my_ret_val)

    return my_ret_val


def get_padding(a_str, a_padding):
    if a_str is None:
        a_str = "-"

    my_retval = str(a_str)

    if len(my_retval) < a_padding:
        my_to_add = a_padding - len(my_retval)

        my_padding = " " * my_to_add
        my_retval = my_retval + my_padding
    else:
        my_retval = my_retval[0:a_padding-2] + "**"

    return my_retval


def get_ago(a_seconds, a_suffix="ago"):
    my_ret_val = "now"

    if a_seconds > 0:
        my_ret_val = "{} secs {}".format(a_seconds, a_suffix)
    if a_seconds > 60:
        my_ret_val = "{} mins {}".format(floor(a_seconds / 60), a_suffix)
    if a_seconds > 86400:
        my_ret_val = "{} days {}".format(floor(a_seconds / 86400), a_suffix)

    return my_ret_val


def get_timestamp_str():
    my_now = datetime.now()

    current_time = my_now.strftime("%H:%M:%S")
    return "{}".format(current_time)


def now(a_mins=0):
    from datetime import datetime, timedelta
    my_time_travel = current_app.config["TIME_TRAVEL"]
    if my_time_travel != '':
        my_ret_val = datetime.strptime(my_time_travel, "%Y-%m-%d %H:%M:%S")
    else:
        my_ret_val = datetime.now()

    if a_mins > 0:
        my_ret_val = my_ret_val - timedelta(minutes=a_mins)

    return my_ret_val


def timestamp_str():
    current_time = now().strftime("%H:%M:%S")
    return "{}".format(current_time)


def parse_time(a_time_str):
    my_ret_val = 0

    try:
        if ":" in a_time_str:
            my_split = a_time_str.split(':')
            my_min = int(my_split[0])
            my_rest = float(my_split[1])
        else:
            my_min = 0
            my_rest = float(a_time_str)

        my_ret_val = my_min * 60 + my_rest
        my_ret_val = round(my_ret_val, 3)
    except Exception as e:
        current_app.logger.error("cannot parse {} to float {}".format(a_time_str, e))

    return my_ret_val


def sec_to_str(a_time):
    my_retval = ""

    my_hr = 0
    if a_time > 3600:
        my_hr = math.floor(a_time/3600)

    my_rest = a_time - my_hr*3600
    my_min = 0
    if my_rest > 60:
        my_min = math.floor(my_rest/60)

    my_sec = round(a_time - my_hr*3600 - my_min*60, 3)

    if my_hr > 0:
        my_retval = "{}h".format(my_hr)

    if my_min > 0:
        my_retval += "{}:{}".format(my_min, my_sec)
    else:
        my_retval += "{}".format(my_sec)

    return my_retval



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app

from .abstract_cmd import AbstractCmd
from ...globals.email_mgr import EmailMgr
from ...db import get_db


class MailTstCmd(AbstractCmd):
    def __init__(self):
        super().__init__("test")

    def action(self, a_param: list):
        my_ret_val = False

        if len(a_param) == 1:
            my_to = a_param[0]
            try:
                my_db_session = get_db().session
                my_email_mgr = EmailMgr(my_db_session)
                my_from = current_app.config["USER_EMAIL_SENDER_EMAIL"]
                my_email_mgr.send_test(my_from, my_to)
                my_db_session.commit()

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: email sent" ]]'
            except Exception as ex:
                my_ret_mesg = '[[ print "Error: problems sending email {}" ]]'.format(ex)
        else:
            my_ret_mesg = self.help()

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: mail t <email>\nsends test email" ]]'


class MailCmd(AbstractCmd):
    def __init__(self):
        super().__init__("mail")
        self.reg_cmd(["t", "test"], MailTstCmd())

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_result, my_mesg = my_cmd.exec(a_params)

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: mail {test}" ]]'




#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

import click
import os
import shutil


@click.group()
def cli():
    """Soseki CLI - Tools for managing Soseki applications"""
    pass


@cli.command('init')
@click.argument('app_name', required=False, default='app')
def init_command(app_name):
    """Initialize a new Soseki application with standard folder structure."""
    
    if os.path.exists(app_name):
        click.echo(f'Error: Directory "{app_name}" already exists')
        return
    
    # Get the path to the blanco_app template
    blanco_app_path = os.path.join(os.path.dirname(__file__), 'blanco_app')
    
    if not os.path.exists(blanco_app_path):
        click.echo(f'Error: Template directory not found at {blanco_app_path}')
        return
    
    # Copy the entire blanco_app directory
    try:
        shutil.copytree(blanco_app_path, app_name)
        click.echo(f'Created: {app_name}/ (from template)')
        
        # Get the parent directory where app_name is being created
        parent_dir = os.path.dirname(os.path.abspath(app_name))
        
        # Copy bin directory next to app folder
        blanco_bin_path = os.path.join(os.path.dirname(__file__), 'blanco_bin')
        bin_dest = os.path.join(parent_dir, 'bin')
        
        if os.path.exists(blanco_bin_path):
            if os.path.exists(bin_dest):
                shutil.rmtree(bin_dest)
            shutil.copytree(blanco_bin_path, bin_dest)
            click.echo(f'Created: bin/ (from template)')
        
        # Copy jup directory next to app folder
        blanco_jup_path = os.path.join(os.path.dirname(__file__), 'blanco_jup')
        jup_dest = os.path.join(parent_dir, 'jup')
        
        if os.path.exists(blanco_jup_path):
            if os.path.exists(jup_dest):
                shutil.rmtree(jup_dest)
            shutil.copytree(blanco_jup_path, jup_dest)
            click.echo(f'Created: jup/ (from template)')
        
        # List created structure
        for root, dirs, files in os.walk(app_name):
            level = root.replace(app_name, '').count(os.sep)
            indent = ' ' * 2 * level
            click.echo(f'{indent}{os.path.basename(root)}/')
            subindent = ' ' * 2 * (level + 1)
            for file in files:
                click.echo(f'{subindent}{file}')
        
        click.echo(f'\nSuccessfully initialized Soseki application in "{app_name}/"')
        click.echo(f'\nNext steps:')
        click.echo(f'  # Edit cfg/lite.yaml with your settings')
        click.echo(f'  ./bin/run_app.sh')
        
    except Exception as e:
        click.echo(f'Error: Failed to create application: {e}')


if __name__ == '__main__':
    cli()



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app

from .abstract_cmd import AbstractCmd
from ...models.user import User
from ...db import get_db, truncate_password


class ChangePasswdCmd(AbstractCmd):
    def __init__(self):
        super().__init__("passwd")

    def has_numbers(self, a_str):
        return any(char.isdigit() for char in a_str)

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = self.help()

        if len(a_params) == 2:
            my_user = User.get_by_email(a_params[0])

            if my_user is not None:
                my_pass = a_params[1]

                my_pass_len = 5
                if len(my_pass) > my_pass_len and self.has_numbers(my_pass):
                    user_manager = current_app.user_manager
                    my_user.password = user_manager.hash_password(truncate_password(my_pass))

                    get_db().session.add(my_user)
                    get_db().session.commit()

                    my_ret_val = True
                    my_ret_mesg = '[[ print "OK: user {} password changed" ]]'.format(a_params[0])
                else:
                    my_ret_mesg = '[[ print "Error: password must be longer than {} and contain numbers" ]]'.format(my_pass_len)
            else:
                my_ret_mesg = '[[ print "Error: user {} not found" ]]'.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: passwd <email> <password>\nchange password" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from flask import current_app

from ssk.logic.cmd.abstract_cmd import AbstractCmd
from ssk.logic.jobs.db_cleanup_job import DbCleanupJob


class DbCleanupCmd(AbstractCmd):
    def __init__(self):
        super().__init__("dbcleanup")

    def action(self, a_params: list):
        from ssk.globals.cmd_processor import CmdProcessor

        my_retval = None

        if len(a_params) == 0:
            if "DB_CLEANUP" in current_app.config.keys():
                my_to_clean = current_app.config["DB_CLEANUP"]
                my_mesg = '[[ print "OK: cleaning up '

                for my_tmp_table in my_to_clean.keys():
                    my_task = DbCleanupJob(current_app, a_args=["dbcleanup {}".format(my_tmp_table), my_tmp_table, my_to_clean[my_tmp_table]])
                    CmdProcessor.submit_cmd(my_task)
                    my_mesg += "{}, ".format(my_tmp_table)
                    my_retval = my_task.get_task_id()

                my_mesg += '" ]]'
            else:
                my_mesg = '[[ print "Error: DB_CLEANUP not set" ]]'
        else:
            my_mesg = '[[ print "Error: action name is missing" ]]'

        return my_retval, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: dbcleanup\nstarts a dbcleanup job" ]]'


#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from ..db import func_db, get_db


class Access(func_db.Model):
    __tablename__ = 'access'
    __bind_key__ = 'logdb'

    id = func_db.Column(func_db.Integer, primary_key=True)
    remote_addr = func_db.Column(func_db.String(100))
    method = func_db.Column(func_db.String(100))
    protocol = func_db.Column(func_db.String(100))
    path = func_db.Column(func_db.String(250))
    response = func_db.Column(func_db.String(100))
    response_time = func_db.Column(func_db.Integer)
    user = func_db.Column(func_db.String(100))

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    def __repr__(self):
        return '<access %s>' % self.path

    @staticmethod
    def get_by_key(a_key):
        my_ret_val = get_db().session.query(Access).filter(Access.req_key == a_key).first()

        return my_ret_val



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app
from sqlalchemy import and_
from ..db import get_db
from ssk.globals.setting_parser import SettingParser


class AppSettings:
    ERROR_KEY_EMPTY = "Key cannot be empty"

    _globals = {}

    def get_setting(self, a_key, an_owners_email=None):
        from ssk.models.all_ssk_db import User
        from ssk.models.all_ssk_db import Setting
        my_ret_val = None

        if an_owners_email is not None:
            my_user = User.get_by_email(an_owners_email)
            if my_user is not None:
                my_setting = Setting.get_by_owner_key(my_user.id, a_key)

                if my_setting is not None:
                    my_ret_val = SettingParser.parse(my_setting.type, my_setting.value)
        else:
            if a_key not in self._globals.keys():
                my_setting = Setting.get_by_key(a_key)

                if my_setting is not None:
                    my_ret_val = SettingParser.parse(my_setting.type, my_setting.value)
                    self._globals[a_key] = my_ret_val
            else:
                my_ret_val = self._globals[a_key]

        return my_ret_val

    def del_setting(self, an_owners_email, a_key):
        from ssk.models.all_ssk_db import User
        from ssk.models.all_ssk_db import Setting

        my_ret_val = False

        my_user = User.get_by_email(an_owners_email)
        if my_user is None:
            return False, "Error: User {} not found".format(an_owners_email)

        if a_key == "":
            return False, AppSettings.ERROR_KEY_EMPTY

        from ..db import get_db
        my_setting = get_db().session.query(Setting).filter(and_(Setting.key == a_key,
                                                                 Setting.user_id == my_user.id)).first()

        if my_setting is not None:
            if not my_setting.system:
                get_db().session.delete(my_setting)
                get_db().session.commit()
                my_ret_val = True
                my_ret_mesg = "OK: {} removed".format(a_key)
            else:
                my_ret_mesg = "Error: Cannot remove system settings"
        else:
            my_ret_mesg = "Error: {} not found".format(a_key)

        return my_ret_val, my_ret_mesg

    def set_global(self, a_key, a_value):
        from ssk.models.all_ssk_db import User
        from ssk.models.all_ssk_db import Setting

        my_ret_val = False

        if a_key == "":
            return False, AppSettings.ERROR_KEY_EMPTY

        my_admin_user = get_db().session.query(User).filter(User.username == current_app.config["ADMIN_NAME"]).first()
        if my_admin_user is not None:
            my_setting = get_db().session.query(Setting).filter(and_(Setting.key == a_key,
                                                                     Setting.user_id == my_admin_user.id)).first()

            if my_setting is not None:
                my_setting.system = a_value
                get_db().session.add(my_setting)
                get_db().session.commit()

                if my_setting.system and a_key not in self._globals.keys():
                    self._globals[a_key] = my_setting.value

                my_ret_val = True
                my_ret_mesg = "OK: {} system changed to {}".format(a_key, a_value)
            else:
                my_ret_mesg = "Error: {} not found".format(a_key)
        else:
            my_ret_mesg = "Error: {} admin user not found".format(a_key)

        return my_ret_val, my_ret_mesg

    def is_global(self, a_key):
        from ssk.models.all_ssk_db import User
        from ssk.models.all_ssk_db import Setting

        my_ret_val = False
        my_ret_mesg = "OK: {} is NOT a system conf".format(a_key)

        if a_key == "":
            return False, AppSettings.ERROR_KEY_EMPTY

        my_admin_user = get_db().session.query(User).filter(User.username == current_app.config["ADMIN_NAME"]).first()
        if my_admin_user is not None:
            my_setting = get_db().session.query(Setting).filter(and_(Setting.key == a_key,
                                                                     Setting.user_id == my_admin_user.id)).first()

            if my_setting:
                if my_setting.system:
                    my_ret_val = True
                    my_ret_mesg = "OK: {} is a system conf".format(a_key)
            else:
                my_ret_mesg = "Error: setting {} not found".format(a_key)
        else:
            my_ret_mesg = "Error: {} admin user not found".format(a_key)

        return my_ret_val, my_ret_mesg

    def set_setting(self, an_owners_email, a_key, a_type, a_value):
        from ssk.models.all_ssk_db import User
        from ssk.models.all_ssk_db import Setting

        my_user = User.get_by_email(an_owners_email)
        if my_user is None:
            return False, "Error: User {} not found".format(an_owners_email)

        if a_key == "":
            return False, AppSettings.ERROR_KEY_EMPTY

        if not SettingParser.is_valid(a_type):
            return False, "Error: Invalid Type {}".format(a_type)

        if a_value == "":
            return False, "Error: Value cannot be empty"

        my_value = SettingParser.parse(a_type, a_value)
        if my_value is None:
            return False, "Error: Value {} cannot be parsed".format(a_value)

        my_setting = get_db().session.query(Setting).filter(and_(Setting.key == a_key,
                                                                 Setting.user_id == my_user.id)).first()

        if my_setting is not None:
            my_old_value = my_setting.value
            my_setting.value = my_value
            my_setting.type = a_type

            get_db().session.add(my_setting)
            get_db().session.commit()

            if my_setting.system:
                self._globals[a_key] = my_value

            my_ret_val = True
            my_ret_mesg = "OK: {} changed from {} to {}".format(a_key, my_old_value, my_value)
        else:
            my_setting = Setting()
            my_setting.key = a_key
            my_setting.type = a_type
            my_setting.value = a_value

            my_user.user_settings.append(my_setting)
            get_db().session.commit()

            my_ret_val = True
            my_ret_mesg = "OK: {} added. Value set to {}".format(a_key, my_value)

        return my_ret_val, my_ret_mesg



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
from setuptools.command.sdist import sdist
import re
import os
import shutil

def sync_blanco_app():
    """Copy app directory to ssk/blanco_app before building."""
    app_dir = 'app'
    blanco_app_dir = 'ssk/blanco_app'
    
    if os.path.exists(app_dir):
        # Remove existing blanco_app if it exists
        if os.path.exists(blanco_app_dir):
            shutil.rmtree(blanco_app_dir)
        
        # Copy app to blanco_app, excluding templates and static
        def ignore_patterns(dir, files):
            # Ignore templates and static directories at the root level
            if dir == app_dir:
                return {'templates', 'static'}
            return set()
        
        shutil.copytree(app_dir, blanco_app_dir, ignore=ignore_patterns)
        print(f"Synced {app_dir}/ to {blanco_app_dir}/ (excluding templates and static)")
    else:
        print(f"Warning: {app_dir}/ directory not found, skipping sync")
    
    # Sync bin directory
    bin_dir = 'bin'
    blanco_bin_dir = 'ssk/blanco_bin'
    
    if os.path.exists(bin_dir):
        if os.path.exists(blanco_bin_dir):
            shutil.rmtree(blanco_bin_dir)
        shutil.copytree(bin_dir, blanco_bin_dir)
        print(f"Synced {bin_dir}/ to {blanco_bin_dir}/")
    else:
        print(f"Warning: {bin_dir}/ directory not found, skipping sync")
    
    # Sync jup directory
    jup_dir = 'jup'
    blanco_jup_dir = 'ssk/blanco_jup'
    
    if os.path.exists(jup_dir):
        if os.path.exists(blanco_jup_dir):
            shutil.rmtree(blanco_jup_dir)
        shutil.copytree(jup_dir, blanco_jup_dir)
        print(f"Synced {jup_dir}/ to {blanco_jup_dir}/")
    else:
        print(f"Warning: {jup_dir}/ directory not found, skipping sync")


class CustomBuildPy(build_py):
    """Custom build command that syncs blanco_app before building."""
    def run(self):
        sync_blanco_app()
        build_py.run(self)


class CustomSdist(sdist):
    """Custom sdist command that syncs blanco_app before creating source distribution."""
    def run(self):
        sync_blanco_app()
        sdist.run(self)


# Read version without importing the module
with open('ssk/ssk_consts.py', 'r') as f:
    version_file = f.read()
    version_match = re.search(r"^SSK_VER = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        SSK_VER = version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")

# Read long description from README
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

# Read requirements
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Filter out comments and empty lines
install_requires = []
for req in requirements:
    req = req.strip()
    if req and not req.startswith('#'):
        install_requires.append(req)

setup(
    name='soseki',
    version=SSK_VER,
    description='a lightweight foundation for building Python web tools.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://codingminds.io/soseki',
    author='Michał Świtała',
    author_email='michal@codingminds.io',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Framework :: Flask',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    keywords='flask web framework user-management scheduling',
    packages=find_packages(exclude=['tests', 'tests.*', 'app', 'app.*', 'scripts', 'jup', 'docs']),
    python_requires='>=3.10',
    install_requires=install_requires,
    package_data={
        'ssk': [
            'templates/**/*',
            'static/**/*',
            'blanco_app/**/*',
            'blanco_bin/**/*',
            'blanco_jup/**/*',
        ]
    },
    include_package_data=True,
    zip_safe=False,
    cmdclass={
        'build_py': CustomBuildPy,
        'sdist': CustomSdist,
    },
    entry_points={
        'console_scripts': [
            'soseki=ssk.cli:cli',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/acodingmind/soseki/issues',
        'Source': 'https://github.com/acodingmind/soseki',
    },
)



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from .base_job import BaseJob
from sqlalchemy import text
from ... import get_db


class DbStatJob(BaseJob):
    def __init__(self, an_app, a_args):
        super(DbStatJob, self).__init__(an_app, a_args)

    def work(self):
        from ssk.models.db_stats import DBStats

        with self._app.app_context():
            try:
                my_all_tables_dict = get_db().metadata.tables.keys()
                my_all_tables = [my_tmp_name for my_tmp_name in my_all_tables_dict]
                my_log_tables_dict = get_db().metadatas['logdb'].tables.keys()
                my_log_tables = [my_tmp_name for my_tmp_name in my_log_tables_dict]
                my_all_tables += my_log_tables

                for my_tmp_table in my_all_tables:
                    my_count_sql = text("select count(*) from {}".format(my_tmp_table))

                    my_prefix = "func"
                    if my_tmp_table not in my_log_tables:
                        my_counter = get_db().session.scalar(my_count_sql)
                    else:
                        my_count_sql = text("select count(*) from {}".format(my_tmp_table))
                        my_prefix = "logdb"

                        my_log_engine = get_db().engines['logdb']
                        my_counter = get_db().session.execute(my_count_sql, None, bind_arguments={'bind': my_log_engine}).scalar()

                    my_db_stat = DBStats()
                    my_db_stat.type = "adhoc"
                    my_db_stat.table = "{}.{}".format(my_prefix, my_tmp_table)
                    my_db_stat.counter = my_counter

                    my_db_session = get_db().session
                    my_db_session.add(my_db_stat)
                    my_db_session.commit()
            except Exception as problem:
                self.write_to_log("Db Stat Failed {}".format(problem))



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from datetime import timedelta

from .base_job import BaseJob
from ... import get_db
from ...utils import now


class DbCleanupJob(BaseJob):
    def __init__(self, an_app, a_args):
        super(DbCleanupJob, self).__init__(an_app, a_args)

    def work(self):
        with self._app.app_context():
            my_args = self.get_args()

            if len(my_args) == 3:
                my_table_name = my_args[1]
                my_days = my_args[2]

                my_since = now() - timedelta(days=my_days)
                self.write_to_log("cleanup {} start {} until {}".format(my_args[0], my_table_name, my_since))

                try:
                    my_engine = None
                    my_table_split = my_table_name.split('.')
                    if len(my_table_split) > 1:
                        my_table_name = my_table_split[1]
                        my_engine = get_db().get_engine(bind_key=my_table_split[0])
                    else:
                        my_table_name = my_table_split[0]
                        my_engine = get_db().engine

                    my_pre = "select count(*) from {} where created < '{}'".format(my_table_name, my_since)
                    my_to_delete = my_engine.scalar(my_pre)
                    my_sql = "delete from {} where created < '{}'".format(my_table_name, my_since)
                    my_engine.execute(my_sql)
                    my_check = my_engine.scalar(my_pre)
                    if my_check == 0:
                        self.write_to_audit("OK", "cleanup table {} until {} ({} records deleted)".format(my_table_name, my_since.strftime("%D %H:%M:%S"), my_to_delete))
                except Exception as problem:
                    self.write_to_log("Db Cleanup {} {} Failed {}".format(my_table_name, my_since, problem))

                self.write_to_log("Db Cleanup Finished successfully")
            else:
                self.write_to_log("Db Cleanup Failed not enough arguments")



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

import inspect
import os
import shutil

from flask import Flask

import ssk
from ssk import init_ssk, start_ssk
from pathlib import Path


def create_app(testing=None):
    my_app_templates = "app/templates"
    my_app_static = "app/static"

    my_ssk_path = os.path.dirname(inspect.getfile(ssk))
    my_ssk_templates = "{}/templates/".format(my_ssk_path)
    my_ssk_static = "{}/static/".format(my_ssk_path)

    if Path(my_app_templates).is_symlink():
        os.unlink(my_app_templates)
    os.symlink(my_ssk_templates, my_app_templates)

    if Path(my_app_static).is_symlink():
        os.unlink(my_app_static)
    os.symlink(my_ssk_static, my_app_static)

    shutil.rmtree("app/templates/local", ignore_errors=True)
    shutil.copytree("app/html/local", "app/templates/local")

    shutil.rmtree("app/static/local", ignore_errors=True)
    shutil.copytree("app/assets/local", "app/static/local")

    my_instance_path = os.path.join(os.getcwd(), 'ins')
    my_app = Flask(__name__, instance_relative_config=True, instance_path=my_instance_path)

    from .logic.app_logic import AppLogic
    from .db_upgrader import DbUpgrader

    my_app = init_ssk(my_app, AppLogic, DbUpgrader, testing)
    start_ssk(my_app, testing)
    my_app.logger.info("create_app {} v. {} soseki ver {} end".format(my_app.config["USER_APP_NAME"],
                                                                       my_app.config["USER_APP_VERSION"],
                                                                       my_app.config["SSK_VER"]))

    from app.blueprints import home
    my_app.register_blueprint(home.bp)

    from app.blueprints import api
    my_app.register_blueprint(api.bp)

    return my_app



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask_login import current_user

from .abstract_cmd import AbstractCmd
from ... import get_db
from ...models.audit import Audit
from ...utils import get_padding


class IntroCmd(AbstractCmd):
    def __init__(self):
        super().__init__("(w)hy", a_type=AbstractCmd.ANON)

    def action(self, a_param: list):
        my_mesg = '[[ print "'
        my_mesg += "\nsoseki.io has a command line interface because cmd can be extended quickly with a minimum risk of regression.\nThis way features can be delivered faster, and extended with a better GUI afterwards when real user feedback is available.\n"
        my_mesg = my_mesg + '" ]]'

        return True, my_mesg

    def help(self, a_wrapped=True):
        my_help = "why cmd in a web based tool ?"
        if a_wrapped:
            my_help = '[[ print "Usage: ?\n{}" ]]'.format(my_help)

        return my_help


class AnonCmd(AbstractCmd):
    def __init__(self):
        super().__init__("anon", a_type=AbstractCmd.ANON)
        self.reg_cmd(["w", "why"], IntroCmd())

    def action(self, a_params):
        my_ret_val = False
        my_ret_mesg = self.help()

        if len(a_params) > 0:
            my_cmd_name = self.get_params(a_params)
            my_cmd = self.get_cmd(my_cmd_name)
            if my_cmd is not None:
                my_audit = Audit()

                my_params_str = " ".join(a_params)
                my_ret_val, my_ret_mesg = my_cmd.exec(a_params)

                my_audit.category = "CMD"
                if current_user.is_anonymous:
                    my_audit.by_user = "anonymous"
                else:
                    my_audit.by_user = current_user.email
                my_audit.description = "{} {}: {}".format(my_cmd_name, my_params_str, my_ret_val)

                if my_ret_val:
                    my_audit.status = "OK"
                else:
                    my_audit.status = "NOK"
                get_db().session.add(my_audit)
                get_db().session.commit()

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        my_names = []
        my_mesg = '[[ print "supported commands:'

        for my_tmp_cmd in self.get_all_cmds().values():
            my_name = my_tmp_cmd.get_name()
            if my_name not in my_names:
                my_mesg += "\n{} : {}".format(get_padding(my_name, 10), my_tmp_cmd.help(False))
                my_names.append(my_name)

        my_mesg += '\n" ]]'

        return my_mesg



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import logging
from ssk.ssk_consts import SSK_VER, SSK_MODEL_VERSION, SSK_ADMIN_GROUP


def get_log_level(an_app):
    my_ret_val = logging.DEBUG
    if an_app.config["LOG_LEVEL"] == "INFO":
        my_ret_val = logging.INFO
    elif an_app.config["LOG_LEVEL"] == "WARN":
        my_ret_val = logging.WARN
    elif an_app.config["LOG_LEVEL"] == "ERROR":
        my_ret_val = logging.ERROR
    elif an_app.config["LOG_LEVEL"] == "FATAL":
        my_ret_val = logging.FATAL
    elif an_app.config["LOG_LEVEL"] == "CRITICAL":
        my_ret_val = logging.CRITICAL

    return my_ret_val


class Config(object):
    SSK_VER = SSK_VER
    SSK_DB_MODEL_VERSION = SSK_MODEL_VERSION

    # user init
    USER_GROUP_NAME = "user"
    ADMIN_GROUP_NAME = SSK_ADMIN_GROUP

    ADMIN_NAME = "admin"
    USER_FREE_NAME = "user_free"



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from threading import Thread
from queue import Queue


class CmdProcessor:
    KILL_CMD = "k"
    _cmd_queue = Queue()
    _cmd_cnt = 0
    _job_mgr = None
    _logger = None
    _workers = []

    @staticmethod
    def log(a_mesg, a_level="INFO"):
        if a_level == "DEBUG":
            CmdProcessor._logger.debug(a_mesg)
        else:
            CmdProcessor._logger.info(a_mesg)

    @staticmethod
    def process_cmd(a_job_mgr, a_num, a_job_queue):
        CmdProcessor.log("starting command worker {}".format(a_num))

        while True:
            CmdProcessor.log("worker {} waiting for command".format(a_num))

            my_cmd = a_job_queue.get()

            if my_cmd == CmdProcessor.KILL_CMD:
                raise SystemExit()

            if a_job_mgr.is_queued(my_cmd.get_task_id()):
                CmdProcessor.log("worker {} processing command {}".format(a_num, my_cmd.get_task_id()))
                CmdProcessor._cmd_cnt += 1
                try:
                    a_job_mgr.start_job(my_cmd)
                except SystemExit:
                    CmdProcessor.log("worker {} processing command {} stopped".format(a_num, my_cmd.get_task_id()))
                except Exception as e:
                    CmdProcessor.log("worker {} processing command {} error {}".format(a_num, my_cmd.get_task_id(), e))

                CmdProcessor.log("worker {} processing command {} ready".format(a_num, my_cmd.get_task_id()))

            a_job_queue.task_done()

    @staticmethod
    def start(a_job_mgr, a_logger, a_max_collectors):
        CmdProcessor._job_mgr = a_job_mgr
        CmdProcessor._logger = a_logger
        CmdProcessor.log("starting")

        for my_tmp_idx in range(a_max_collectors):
            worker = Thread(target=CmdProcessor.process_cmd, args=(a_job_mgr, my_tmp_idx, CmdProcessor._cmd_queue))
            CmdProcessor._workers.append(worker)
            worker.start()

    @staticmethod
    def stop():
        for my_tmp_idx in range(len(CmdProcessor._workers)):
            CmdProcessor.log("sending kill to worker {}".format(my_tmp_idx))
            CmdProcessor._cmd_queue.put(CmdProcessor.KILL_CMD)

    @staticmethod
    def submit_cmd(a_cmd):
        CmdProcessor._job_mgr.queue_job(a_cmd)
        CmdProcessor._cmd_queue.put(a_cmd)

        return CmdProcessor._cmd_queue.qsize()

    @staticmethod
    def cmd_queue_len():
        return CmdProcessor._cmd_queue.qsize()

    @staticmethod
    def data_processed():
        return CmdProcessor._cmd_cnt



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app


class SettingParser:
    STR_TYPE = "string"
    INT_TYPE = "int"
    FLT_TYPE = "float"
    BOOL_TYPE = "bool"
    JSON_TYPE = "json"

    ERROR_KEY_EMPTY = "Key cannot be empty"

    _valid_types = [STR_TYPE, INT_TYPE, FLT_TYPE, BOOL_TYPE, JSON_TYPE]
    _globals = {}

    @staticmethod
    def is_valid(a_type):
        return a_type in SettingParser._valid_types

    @staticmethod
    def parse(a_type, a_value):
        my_ret_val = None

        try:
            if a_type == SettingParser.STR_TYPE:
                my_ret_val = str(a_value)
            elif a_type == SettingParser.INT_TYPE:
                my_ret_val = int(a_value)
            elif a_type == SettingParser.FLT_TYPE:
                my_ret_val = float(a_value)
            elif a_type == SettingParser.BOOL_TYPE:
                my_ret_val = a_value is True or (a_value == "True") or (a_value == '1')
            elif a_type == SettingParser.JSON_TYPE:
                my_ret_val = a_value
            else:
                current_app.logger.error("Unknown setting type {} for {}".format(a_type, a_value))
        except Exception:
            my_mesg = "Cannot parse type {} for {}".format(a_type, a_value)
            current_app.logger.error(my_mesg)

        return my_ret_val
    


#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from sqlalchemy import and_, desc

from ..db import func_db, get_db


class Stats(func_db.Model):
    __tablename__ = 'stats'
    __bind_key__ = 'logdb'

    id = func_db.Column(func_db.Integer, primary_key=True)
    day = func_db.Column(func_db.String(16))
    path = func_db.Column(func_db.String(10))
    hits = func_db.Column(func_db.Integer)
    mean_response_time = func_db.Column(func_db.Float)
    max_response_time = func_db.Column(func_db.Integer)
    last_processed = func_db.Column(func_db.Integer)

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    def __repr__(self):
        return '<stats %s>' % self.path

    @staticmethod
    def get_last_processed_id():
        my_ret_val = 0
        my_stats_db = get_db().session.query(Stats).order_by(desc(Stats.created)).first()
        if my_stats_db is not None:
            my_ret_val = my_stats_db.last_processed

        return my_ret_val

    @staticmethod
    def get_all():
        my_ret_val = get_db().session.query(Stats).order_by(desc(Stats.created)).all()

        return my_ret_val

    @staticmethod
    def get_last(a_days):
        my_ret_val = get_db().session.query(Stats).order_by(desc(Stats.created)).limit(a_days)

        return my_ret_val

    @staticmethod
    def get_since(a_date):
        my_ret_val = get_db().session.query(Stats).filter(Stats.created > a_date).order_by(desc(Stats.created)).all()

        return my_ret_val

    @staticmethod
    def get_by_date(a_day, a_page):
        my_ret_val = get_db().session.query(Stats).filter(and_(Stats.day == a_day, Stats.path == a_page)).first()

        return my_ret_val
